package ims.data;

import ims.util.Util;

import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public abstract class AbstractSentence implements ISentence {

	protected final int sentId;
	protected final String[] id;
	protected final String[] form;
	
	protected AbstractSentence(int sentId,String[] id,String[] form){
		this.sentId=sentId;
		this.id=id;
		this.form=form;
	}
	
	public int getSentId(){
		return sentId;
	}
	
	public String[] getForms(){
		return form;
	}
	
	@Override
	public int getLength() {
		return form.length;
	}
	
	@Override
	public void reverseSentence(){
		reverseObj(id);
		reverseObj(form);
	}
	
	@Override
	public String getTabbedToString() {
		List<String> l=getTabbedLineList();
		return Util.joinStrings(l, "\n");
	}
	
	public static void reverseHeads(int[] heads){
		//Recompute new head
		for(int i=1;i<heads.length;++i){
			if(heads[i]>0)
				heads[i]=heads.length-heads[i];
		}
		//Then swap them
		for(int i=1,t=heads.length-1;i<t;++i,--t){
			int tmp=heads[i];
			heads[i]=heads[t];
			heads[t]=tmp;
		}
	}
	
	public static <T> void reverseObj(T[] s){
		//Here swap contents of
		//1 with length-1
		//2 with length-2
		//...
		for(int i=1,t=s.length-1;i<t;++i,--t){
			T tmp=s[i];
			s[i]=s[t];
			s[t]=tmp;
		}
	}
	
	
	private static final Pattern SIMPLE_ID_PATTERN=Pattern.compile("\\d+");
	private static final Pattern SENT_ID_TOK_ID_PATTERN=Pattern.compile("(\\d+)_(\\d+)");
	protected void insertTokenOffsetInID(int tokOffset){
		if(tokOffset==0)
			return;
		for(int i=1;i<id.length;++i){
			if(SIMPLE_ID_PATTERN.matcher(id[i]).matches()){
				int x=Integer.parseInt(id[i]);
				id[i]=Integer.toString(x+tokOffset);
			} else {
				Matcher m=SENT_ID_TOK_ID_PATTERN.matcher(id[i]);
				if(m.matches()){
					String sId=m.group(1);
					int tId=Integer.parseInt(m.group(2));
					id[i]=sId+"_"+Integer.toString(tId+tokOffset);
				} //else do nothing, leave id as is
			}
		}
	}
	
	public void renormalizeTokenOffsetInID(int tokOffset) {
		if(tokOffset==0)
			return;
		for(int i=1;i<id.length;++i){
			if(SIMPLE_ID_PATTERN.matcher(id[i]).matches()){
				int x=Integer.parseInt(id[i]);
				id[i]=Integer.toString(x-tokOffset);
			} else {
				Matcher m=SENT_ID_TOK_ID_PATTERN.matcher(id[i]);
				if(m.matches()){
					String sId=m.group(1);
					int tId=Integer.parseInt(m.group(2));
					id[i]=sId+"_"+Integer.toString(tId-tokOffset);
				} //else do nothing
			}
		}
	}
	
	public static void offsetHeadIntIndex(int[] arr,int tokOffset){
		if(tokOffset==0)
			return;
		for(int i=0;i<arr.length;++i){
//			if(arr[i]>0){
			if(arr[i]!=0){
				arr[i]+=tokOffset;
			}
		}
	}

	public static void renormalizeHeadIntIndex(int[] arr,int tokOffset){
		offsetHeadIntIndex(arr,-tokOffset);
	}
}
