package ims.data;

import ims.oracle.IConfiguration;
import gnu.trove.list.TIntList;
import gnu.trove.list.linked.TIntLinkedList;

public class Buffer implements Comparable<Buffer> {

	private final int last;
	private int nxt=1;
	private final TIntList pushedBack;
	
	public Buffer(int last){ //last is the number of the last token (0-indexed at root) -- ``I like apples'' has 3 as last
		this.last=last;
		this.pushedBack=new TIntLinkedList();
	}
	
	private Buffer(int last,int nxt,TIntList pushedBack){
		this.last=last;
		this.nxt=nxt;
		this.pushedBack=pushedBack;
	}
	
	public Buffer copy(){
		Buffer b=new Buffer(last);
		b.nxt=nxt;
		b.pushedBack.addAll(this.pushedBack);
		return b;
	}
	
	public boolean equals(Buffer o){
		if(nxt==o.nxt && last==o.last && this.pushedBack.equals(o.pushedBack))
			return true;
		else
			return false;
	}
	
	public int size(){
		return last-nxt+1+pushedBack.size(); //last=3 for ``I like apples''. With nxt=1 (i.e., just at start), there are 3=3-1+1 tokens left. 
	}
	
	public boolean isEmpty(){
		return size()==0;
	}
	
	public int peekFront(){
		return get(0);
	}

	
	public int get(int n){
		if(pushedBack.isEmpty())
			return nxt+n;
		else if(n<pushedBack.size())
			return pushedBack.get(n);
		else
			return nxt+(n-pushedBack.size());
	}
	
	public int removeFront(){
		if(pushedBack.isEmpty())
			return nxt++;
		else
			return pushedBack.removeAt(0);
	}
		
	public void pushBack(int i){
		if(pushedBack.isEmpty())
			pushedBack.add(i);
		else
			pushedBack.insert(0, i);
	}
	
	@Override
	public int compareTo(Buffer o) {
		int s=this.size();
		int oS=o.size();
		//Fewer items means it's smaller
		if(s<oS)
			return -1;
		else if(oS<s)
			return 1;

		//With same number of items, check the stuff that has been pushed back plus one
		int q=1+Math.max(pushedBack.size(), o.pushedBack.size()); //Have to look this far into the buffer, if they're all the same, it's clear
		//But sometimes this is *too* far, if we are very close to the end
		q=Math.min(q, s);
		for(int r=0;r<q;++r){
			int a=this.get(r);
			int b=o.get(r);
			if(a<b)
				return -1;
			else if(b<a)
				return 1;
		}
		//Otherwise they have to be identical
		return 0;
	}
	
	public StringBuilder appendSignature(StringBuilder sb){
		sb.append((char)size());
		for(int i=0;i<size();++i)
			sb.append((char) get(i));
		return sb;
	}

	public String toString(){
		if(size()==0)
			return "[ ]";
		else if(size()==1)
			return "["+peekFront()+"]";
		else if(pushedBack.isEmpty())
			return "["+nxt+".."+last+"]";
		else {
			StringBuilder sb=new StringBuilder();
			for(int q=0;q<pushedBack.size();++q)
				sb.append(pushedBack.get(q)).append(", ");
			sb.append(nxt);
			if(nxt<last)
				sb.append("..").append(last);
			return sb.toString();
		}
	}

	public String toString(String[] t) {
		if(size()<=0)
			return "[ ]";
		else if(size()==1)
			return "["+t[peekFront()]+"]";
		else if(pushedBack.isEmpty())
			return "["+t[nxt]+" .. "+t[last]+"]";
		else {
			StringBuilder sb=new StringBuilder();
			for(int q=0;q<pushedBack.size();++q)
				sb.append(t[pushedBack.get(q)]).append(", ");
			sb.append(t[nxt]);
			if(nxt<last)
				sb.append(" .. ").append(t[last]);
			return sb.toString();
		}
	}

	public int getBufferNoPushedBack(){
		return pushedBack.size();
	}

	public int getLast() {
		return last;
	}
	
	public static Buffer copyOfBufferShiftedOne(IConfiguration c) {
		int last=c.sentenceSize()-1;
		TIntList pb=new TIntLinkedList();
		int m=c.getBufferNoPushedBack();
		int i=1;
		for(;i<m;++i)
			pb.add(c.peekBuffer(i));
		int next=c.peekBuffer(i);
		if(next==-1)
			next=last+1;
		return new Buffer(last,next,pb);
	}

}
