package ims.data;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

public class NBestCoNLL09Sentence extends CoNLL09Sentence {

	private final List<Parse> parses;
	
	protected NBestCoNLL09Sentence(CoNLL09Sentence s,List<Parse> parses){
		super(s.sentId, s.id, s.form, s.lemma, s.plemma, s.pos, s.ppos, s.feat, s.pfeat, s.head, s.lbl, s.phead, s.plbl, s.rest);
		this.parses=parses;
	}

	public static NBestCoNLL09Sentence lines2Sentence(List<String> lines,Integer sid) {
		List<String> conll09=lines.subList(0, lines.size()-1);
		String nBest=lines.get(lines.size()-1);
		CoNLL09Sentence s=CoNLL09Sentence.lines2Sentence(conll09, sid);
		List<Parse> parses=nBestLineToParseList(nBest);
		return new NBestCoNLL09Sentence(s,parses);
	}
	
	@Override
	public List<String> getTabbedLineList() {
		List<String> conll09=super.getTabbedLineList();
		String nBest=nBestToString(parses);
		conll09.add(nBest);
		return conll09;
	}
	
	@Override
	public void reverseSentence() {
		super.reverseSentence();
		for(Parse p:parses)
			p.reverseParse();
	}
	
	@Override
	public void insertTokenOffsetsInHeadAndIDCols(int tokOffset) {
		throw new Error("not implemented"); //TODO
	}
	
	@Override
	public void renormalizeTokenOffsetsInHeadAndIDCols(int tokOffset) {
		throw new Error("not implemented"); //TODO
	}
	
	@Override
	public CoNLL09Sentence getSubSentence(int b, int e, int tokOffset) {
		throw new Error("not implemented"); //TODO
	}
	
	public StagStackSentence toStagStackPredInStack(){
		throw new Error("not implemented"); //TODO
	}
	
	public static List<Parse> nBestLineToParseList(String line){
		if(!line.startsWith("ParseNBest"))
			throw new Error("malformed ParseNBest line");
		
		String[] fields=line.split("\t");
		String[] treesCompact=fields[2].split("\\];\\[");
		if(fields.length!=3 || treesCompact.length!=Integer.parseInt(fields[1]))
			throw new Error("failed to parse nbest line: "+line);

		treesCompact[0]=treesCompact[0].replaceFirst("\\[", "");
		treesCompact[treesCompact.length-1]=treesCompact[treesCompact.length-1].replaceFirst("\\]$", "");

		List<Parse> parses=new ArrayList<>();
		for(int k=0;k<treesCompact.length;++k)
			parses.add(string2LabeledParse(treesCompact[k]));

		return parses;
	}
	
	private static Parse string2LabeledParse(String s) {
		String[] f=s.split(" @@@ ");
		Map<String,String> parseProperties=recoverProperties(f[0]);
		String[] a=f[1].split(" ");
		int[] heads=new int[a.length];
		String[] lbls=new String[a.length];
		for(int i=0;i<a.length;++i){
			String[] b=a[i].split(",");
			heads[i]=Integer.parseInt(b[0]);
			lbls[i]=b[1];
		}
		return new Parse(heads, lbls, parseProperties);
	}
	
	private static Map<String, String> recoverProperties(String propString) {
		if(propString.length()==0)
			return Collections.emptyMap();
		String[] kvPairs=propString.split(" ");
		Map<String,String> m=new HashMap<>();
		for(String kvPair:kvPairs){
			String[] a=kvPair.split("=",2);
			m.put(a[0], a[1]);
		}
		return m;
	}

	public static String nBestToString(List<Parse> parses){
		StringBuilder sb=new StringBuilder("ParseNBest\t");
		sb.append(parses.size()).append("\t");
		for(Iterator<Parse> it=parses.iterator();it.hasNext();){
			Parse p=it.next();
			appendParseNBestString(sb,p);
			if(it.hasNext())
				sb.append(";");
		}
		return sb.toString();
	}

	private static void appendParseNBestString(StringBuilder sb, Parse p) {
		sb.append("[");
		List<String> keys=new ArrayList<>(p.properties.keySet());
		Collections.sort(keys);
		for(String key:keys)
			sb.append(key).append('=').append(p.properties.get(key)).append(' ');
		if(keys.size()==0)
			sb.append(' ');
		sb.append("@@@ ");
		for(int i=0;i<p.heads.length-1;++i)
			sb.append(p.heads[i]).append(',').append(p.lbls[i]).append(' ');
		sb.append(p.heads[p.heads.length-1]).append(',').append(p.lbls[p.lbls.length-1]).append(']');
	}
	
	public List<Parse> getParses(){
		return parses;
	}

	public void setGoldHeads(int[] heads) {
		System.arraycopy(heads, 0, this.head, 0, heads.length);
	}

	public void setGoldLabels(String[] labels) {
		System.arraycopy(labels, 0, this.lbl, 0, labels.length);
	}
	
}
