package ims.data;

import ims.trans.Options;

import java.util.ArrayList;
import java.util.List;

public class StagStackSentence extends CoNLL09Sentence {

	public String[] stag;
	public int[] stackHead;
	public String[] stackLbl;
	
	protected StagStackSentence(int sentId, String[] ids, String[] forms,String[] lemma, String[] plemma, String[] pos, String[] ppos,String[] feat, String[] pfeat, int[] head, String[] lbl,int[] phead, String[] plbl, String[] stag,int[] stackHead,String[] stackLbl) {
		super(sentId, ids, forms, lemma, plemma, pos, ppos, feat, pfeat, head, lbl,	phead, plbl, null);
		this.stag=stag;
		this.stackHead=stackHead;
		this.stackLbl=stackLbl;
	}

	
	@Override
	public List<String> getTabbedLineList() {
		List<String> lines=new ArrayList<>();
		for(int i=1;i<form.length;++i){
			StringBuilder sb=new StringBuilder();
			sb.append(id[i]).append('\t');
			sb.append(form[i]).append('\t');
			sb.append(lemma[i]).append('\t');
			sb.append(plemma[i]).append('\t');
			sb.append(pos[i]).append('\t');
			sb.append(ppos[i]).append('\t');
			sb.append(feat[i]).append('\t');
			sb.append(pfeat[i]).append('\t');
			if(head[i]<0) sb.append("_").append('\t');
			else sb.append(head[i]).append('\t');
			if(phead[i]<0) sb.append("_").append('\t');
			else sb.append(phead[i]).append('\t');
			sb.append(lbl[i]).append('\t');
			sb.append(plbl[i]).append('\t');
			sb.append(stag[i]).append('\t');
			if(stackHead[i]<0) sb.append("_").append('\t');
			else sb.append(stackHead[i]).append('\t');
			sb.append(stackLbl[i]);			
			lines.add(sb.toString());
		}
		return lines;
	}
	
	public static StagStackSentence lines2Sentence(List<String> lines,Integer sid) {
		int size=lines.size()+1; //one for root
		String[] id=new String[size];
		String[] form=new String[size];
		String[] lemma=new String[size];
		String[] plemma=new String[size];
		String[] pos=new String[size];
		String[] ppos=new String[size];
		String[] feat=new String[size];
		String[] pfeat=new String[size];
		int[] head=new int[size];
		String[] lbl=new String[size];
		int[] pHead=new int[size];
		String[] pLbl=new String[size];
		String[] stag=new String[size];
		int[] stackHead=new int[size];
		String[] stackLbl=new String[size];
		
		form[0]=ISentence.ROOT;
		lemma[0]=ISentence.ROOT;
		plemma[0]=ISentence.ROOT;
		pos[0]=ISentence.ROOT;
		ppos[0]=ISentence.ROOT;
		feat[0]=ISentence.ROOT;
		pfeat[0]=ISentence.ROOT;
		head[0]=-1;
		lbl[0]=ISentence.ROOT;
		pHead[0]=-1;
		pLbl[0]=ISentence.ROOT;
		stag[0]=ISentence.ROOT_STAG;
		stackHead[0]=-1;
		stackLbl[0]=ISentence.ROOT;
		
		int i=1;
		for(String l:lines){
//			System.out.println(l);
			String[] cols=l.split(Options.columnSeparatorRegExp);
			id[i]=cols[0];
			form[i]=cols[1];
			lemma[i]=cols[2];
			plemma[i]=cols[3];
			pos[i]=cols[4];
			ppos[i]=cols[5];
			feat[i]=cols[6];
			pfeat[i]=cols[7];
			head[i]=cols[8].equals("_")?-1:Integer.parseInt(cols[8]);
			pHead[i]=cols[9].equals("_")?-1:Integer.parseInt(cols[9]);
			lbl[i]=cols[10];
			pLbl[i]=cols[11];
			stag[i]=cols[12];
			stackHead[i]=cols[13].equals("_")?-1:Integer.parseInt(cols[13]);
			stackLbl[i]=cols[14];
			++i;
		}
		
		StagStackSentence s=new StagStackSentence(sid,id,form,lemma,plemma,pos,ppos,feat,pfeat,head,lbl,pHead,pLbl,stag,stackHead,stackLbl);
		return s;
	}

	@Override
	public String[] getStag() {
		return stag;
	}
	
	@Override
	public int[] getStackHead() {
		return stackHead;
	}


	@Override
	public String[] getStackLbl() {
		return stackLbl;
	}
	
	public void setStag(String[] stag) {
		this.stag=stag;
	}
	
	public void setStackHead(int[] heads){
		this.stackHead=heads;
	}
	
	public void setStackLbl(String[] lbls){
		this.stackLbl=lbls;
	}

	@Override
	public void reverseSentence() {
		super.reverseSentence();
		reverseObj(stag);
		reverseObj(stackLbl);
		reverseHeads(stackHead);
	}

	@Override
	public void insertTokenOffsetsInHeadAndIDCols(int tokOffset) {
		super.insertTokenOffsetsInHeadAndIDCols(tokOffset);
		super.offsetHeadIntIndex(stackHead, tokOffset);
	}

	@Override
	public void renormalizeTokenOffsetsInHeadAndIDCols(int tokOffset) {
		super.renormalizeTokenOffsetsInHeadAndIDCols(tokOffset);
		super.renormalizeHeadIntIndex(stackHead, tokOffset);
	}
	
}
