package ims.headfinder;

import ims.data.Instance;
import ims.hotcoref.symbols.AbstractSymbolMapping;
import ims.hotcoref.symbols.SymbolTable;
import ims.trans.FeatureExtractor;

public class FE {

	public static final int SINK_LEN=1000;
	public static final int NONE=AbstractSymbolMapping.NONE_IDX;
	public static final int[] NONE_ARR=FeatureExtractor.NONE_ARR;

	protected final int
	bF, //Bits for form and lemma
	bT, //Bits for pos tag
	bM, //Bits for morphology
	bFF, //Bits for full morph tags
	bD; //Bits for distance buckets

	public FE(SymbolTable symTab){
		this.bF=symTab.forms.getBits();
		this.bT=symTab.tags.getBits();
		this.bM=symTab.feats.getBits();
		this.bFF=symTab.ffeats.getBits();
		this.bD=BITS_BUCKETS;
	}

	public static final int BITS_BUCKETS=5;
	public static int getIntBucket(int i){
		if(i<0)
			return 0;
		if(i==0)
			return 1;
		if(i==1)
			return 2;
		if(i==2)
			return 3;
		if(i==3)
			return 4;
		if(i<6)
			return 5;
		if(i<10)
			return 6;
		if(i<20)
			return 7;
		return 8;
	}


	public int extractFeatures(Instance is,int head, int dep,long[] sink){

		final int dist=getIntBucket(Math.abs(head-dep));

		//Stuff on head
		final int hForm=is.form[head];
		final int hPos=is.tag[head];
		final int hLemma=is.lemma[head];
		final int[] hMorph=is.feat[head];

		final int hP1Form,hP1Pos;
//		final int[] hP1Morph;

//		final int hP2Form,hP2Pos;
//		final int[] hP2Morph;

		final int hM1Form,hM1Pos;
//		final int[] hM1Morph;

//		final int hM2Form,hM2Pos;
//		final int[] hM2Morph;

		if(head+1<is.form.length){ //Has head+1
			hP1Form=is.form[head+1];
			hP1Pos=is.tag[head+1];
//			hP1Morph=is.feat[head+1];
		} else { //No head+1
			hP1Form=NONE;
			hP1Pos=NONE;
//			hP1Morph=NONE_ARR;
		}

		if(head-1>=0){ //Has head-1
			hM1Form=is.form[head-1];
			hM1Pos=is.tag[head-1];
//			hM1Morph=is.feat[head-1];
		} else { //No head-1
			hM1Form=NONE;
			hM1Pos=NONE;
//			hM1Morph=NONE_ARR;			
		}


		//Stuff on dep
		final int dForm=is.form[dep];
		final int dPos=is.tag[dep];
		final int dLemma=is.lemma[dep];
		final int[] dMorph=is.feat[dep];

		final int dP1Form,dP1Pos;
//		final int[] dP1Morph;

//		final int dP2Form,dP2Pos;
//		final int[] dP2Morph;

		final int dM1Form,dM1Pos;
//		final int[] dM1Morph;

//		final int dM2Form,dM2Pos;
//		final int[] dM2Morph;

		if(dep+1<is.form.length){ //Has head+1
			dP1Form=is.form[dep+1];
			dP1Pos=is.tag[dep+1];
//			dP1Morph=is.feat[dep+1];
		} else { //No head+1
			dP1Form=NONE;
			dP1Pos=NONE;
//			dP1Morph=NONE_ARR;
		}

		if(dep-1>=0){ //Has head-1
			dM1Form=is.form[dep-1];
			dM1Pos=is.tag[dep-1];
//			dM1Morph=is.feat[dep-1];
		} else { //No head-1
			dM1Form=NONE;
			dM1Pos=NONE;
//			dM1Morph=NONE_ARR;			
		}


		int t=1;
		int i=0;

		//Trivial head/dep bigrams
		sink[i]=t; sink[i]<<=bF; sink[i]|=hForm;  sink[i]<<=bF; sink[i]|=dForm;  ++i; ++t;
		sink[i]=t; sink[i]<<=bF; sink[i]|=hForm;  sink[i]<<=bF; sink[i]|=dLemma; ++i; ++t;
		sink[i]=t; sink[i]<<=bF; sink[i]|=hLemma; sink[i]<<=bF; sink[i]|=dForm;  ++i; ++t;
		sink[i]=t; sink[i]<<=bF; sink[i]|=hLemma; sink[i]<<=bF; sink[i]|=dLemma; ++i; ++t;

		sink[i]=t; sink[i]<<=bF; sink[i]|=hForm;  sink[i]<<=bT; sink[i]|=dPos;   ++i; ++t;
		sink[i]=t; sink[i]<<=bF; sink[i]|=hLemma; sink[i]<<=bT; sink[i]|=dPos;   ++i; ++t;

		sink[i]=t; sink[i]<<=bT; sink[i]|=hPos;   sink[i]<<=bF; sink[i]|=dForm;  ++i; ++t;
		sink[i]=t; sink[i]<<=bT; sink[i]|=hPos;   sink[i]<<=bF; sink[i]|=dLemma; ++i; ++t;

		sink[i]=t; sink[i]<<=bT; sink[i]|=hPos;   sink[i]<<=bT; sink[i]|=dPos;   ++i; ++t;

		//head+dep morph+tag bigrams
		for(int h=0;h<hMorph.length;++h){
			for(int d=0;d<dMorph.length;++d){
				sink[i]=t; sink[i]<<=bT; sink[i]|=hPos; sink[i]<<=bT; sink[i]|=dPos; sink[i]<<=bM; sink[i]|=dMorph[d]; sink[i]<<=bM; sink[i]|=hMorph[h]; ++i; 
			}
		}
		++t;

		//Distance + head/dep
		sink[i]=t; sink[i]<<=bT; sink[i]|=hPos;   sink[i]<<=bT; sink[i]|=dPos;   sink[i]<<=bD; sink[i]|=dist; ++i; ++t;
		sink[i]=t; sink[i]<<=bF; sink[i]|=hForm;  sink[i]<<=bF; sink[i]|=dForm;  sink[i]<<=bD; sink[i]|=dist; ++i; ++t;
		sink[i]=t; sink[i]<<=bF; sink[i]|=hForm;  sink[i]<<=bT; sink[i]|=dPos;   sink[i]<<=bD; sink[i]|=dist; ++i; ++t;
		sink[i]=t; sink[i]<<=bT; sink[i]|=hPos;   sink[i]<<=bF; sink[i]|=dForm;  sink[i]<<=bD; sink[i]|=dist; ++i; ++t;

		//in between pos and forms (and head+dep pos)
//		int FM=t+1+(Options.x0?1:0);
		if(dist>0){
			for(int n=Math.min(head, dep)+1,max=Math.max(head, dep);n<max;++n){
				sink[i]=t;   sink[i]<<=bT; sink[i]|=hPos; sink[i]<<=bT; sink[i]|=dPos; sink[i]<<=bT; sink[i]|=is.tag[n]; ++i;
//				if(Options.x0){
//					sink[i]=t+1; sink[i]<<=bT; sink[i]|=hPos; sink[i]<<=bT; sink[i]|=dPos; sink[i]<<=bF; sink[i]|=is.form[n]; ++i;
//				}
//				if(Options.x9){
//					sink[i]=FM;  sink[i]<<=bT; sink[i]|=hPos; sink[i]<<=bT; sink[i]|=dPos; sink[i]<<=bFF; sink[i]|=is.ffeat[n]; ++i;
//				}
//				if(Options.x9){ //in between morph
//					for(int m:is.feat[n]){
//						sink[i]=FM;  sink[i]<<=bT; sink[i]|=hPos; sink[i]<<=bT; sink[i]|=dPos; sink[i]<<=bM; sink[i]|=m; ++i;
//					}
//				}
			}
		}
		++t;
//		if(Options.x0){
//			++t;
//		}
//
//		if(Options.x9){
//			++t;
//		}
		
//		if(Options.x1){
//			int hFFeat=is.ffeat[head];
//			int dFFeat=is.ffeat[dep];
//			sink[i]=t; sink[i]<<=bT; sink[i]|=hPos; sink[i]<<=bT; sink[i]|=dPos; sink[i]<<=bFF; sink[i]|=hFFeat; sink[i]<<=bFF; sink[i]|=dFFeat; ++t; ++i; 
//		}

		sink[i]=t; sink[i]<<=bT; sink[i]|=hPos;   sink[i]<<=bF; sink[i]|=hP1Form;  sink[i]<<=bT; sink[i]|=dPos; ++i; ++t;
		sink[i]=t; sink[i]<<=bT; sink[i]|=hPos;   sink[i]<<=bF; sink[i]|=hM1Form;  sink[i]<<=bT; sink[i]|=dPos; ++i; ++t;
		sink[i]=t; sink[i]<<=bT; sink[i]|=hPos;   sink[i]<<=bT; sink[i]|=hP1Pos;   sink[i]<<=bT; sink[i]|=dPos; ++i; ++t;
		sink[i]=t; sink[i]<<=bT; sink[i]|=hPos;   sink[i]<<=bT; sink[i]|=hM1Pos;   sink[i]<<=bT; sink[i]|=dPos; ++i; ++t;

		sink[i]=t; sink[i]<<=bT; sink[i]|=dPos;   sink[i]<<=bF; sink[i]|=dP1Form;  sink[i]<<=bT; sink[i]|=hPos; ++i; ++t;
		sink[i]=t; sink[i]<<=bT; sink[i]|=dPos;   sink[i]<<=bF; sink[i]|=dM1Form;  sink[i]<<=bT; sink[i]|=hPos; ++i; ++t;
		sink[i]=t; sink[i]<<=bT; sink[i]|=dPos;   sink[i]<<=bT; sink[i]|=dP1Pos;   sink[i]<<=bT; sink[i]|=hPos; ++i; ++t;
		sink[i]=t; sink[i]<<=bT; sink[i]|=dPos;   sink[i]<<=bT; sink[i]|=dM1Pos;   sink[i]<<=bT; sink[i]|=hPos; ++i; ++t;


//		if(Options.x2){
//			sink[i]=t; sink[i]<<=bF; sink[i]|=hForm;   sink[i]<<=bF; sink[i]|=hP1Form;  sink[i]<<=bT; sink[i]|=dPos; ++i; ++t;
//			sink[i]=t; sink[i]<<=bF; sink[i]|=hForm;   sink[i]<<=bF; sink[i]|=hM1Form;  sink[i]<<=bT; sink[i]|=dPos; ++i; ++t;
//			sink[i]=t; sink[i]<<=bF; sink[i]|=hForm;   sink[i]<<=bT; sink[i]|=hP1Pos;   sink[i]<<=bT; sink[i]|=dPos; ++i; ++t;
//			sink[i]=t; sink[i]<<=bF; sink[i]|=hForm;   sink[i]<<=bT; sink[i]|=hM1Pos;   sink[i]<<=bT; sink[i]|=dPos; ++i; ++t;
//
//			sink[i]=t; sink[i]<<=bF; sink[i]|=dForm;   sink[i]<<=bF; sink[i]|=dP1Form;  sink[i]<<=bT; sink[i]|=hPos; ++i; ++t;
//			sink[i]=t; sink[i]<<=bF; sink[i]|=dForm;   sink[i]<<=bF; sink[i]|=dM1Form;  sink[i]<<=bT; sink[i]|=hPos; ++i; ++t;
//			sink[i]=t; sink[i]<<=bF; sink[i]|=dForm;   sink[i]<<=bT; sink[i]|=dP1Pos;   sink[i]<<=bT; sink[i]|=hPos; ++i; ++t;
//			sink[i]=t; sink[i]<<=bF; sink[i]|=dForm;   sink[i]<<=bT; sink[i]|=dM1Pos;   sink[i]<<=bT; sink[i]|=hPos; ++i; ++t;
//		}
//
//		if(Options.x3){
//			sink[i]=t; sink[i]<<=bF; sink[i]|=hForm;   sink[i]<<=bF; sink[i]|=hP1Form;  sink[i]<<=bF; sink[i]|=dForm; ++i; ++t;
//			sink[i]=t; sink[i]<<=bF; sink[i]|=hForm;   sink[i]<<=bF; sink[i]|=hM1Form;  sink[i]<<=bF; sink[i]|=dForm; ++i; ++t;
//			sink[i]=t; sink[i]<<=bF; sink[i]|=hForm;   sink[i]<<=bT; sink[i]|=hP1Pos;   sink[i]<<=bF; sink[i]|=dForm; ++i; ++t;
//			sink[i]=t; sink[i]<<=bF; sink[i]|=hForm;   sink[i]<<=bT; sink[i]|=hM1Pos;   sink[i]<<=bF; sink[i]|=dForm; ++i; ++t;
//
//			sink[i]=t; sink[i]<<=bF; sink[i]|=dForm;   sink[i]<<=bF; sink[i]|=dP1Form;  sink[i]<<=bF; sink[i]|=hForm; ++i; ++t;
//			sink[i]=t; sink[i]<<=bF; sink[i]|=dForm;   sink[i]<<=bF; sink[i]|=dM1Form;  sink[i]<<=bF; sink[i]|=hForm; ++i; ++t;
//			sink[i]=t; sink[i]<<=bF; sink[i]|=dForm;   sink[i]<<=bT; sink[i]|=dP1Pos;   sink[i]<<=bF; sink[i]|=hForm; ++i; ++t;
//			sink[i]=t; sink[i]<<=bF; sink[i]|=dForm;   sink[i]<<=bT; sink[i]|=dM1Pos;   sink[i]<<=bF; sink[i]|=hForm; ++i; ++t;
//		}
//
//		if(Options.x4){
//			sink[i]=t; sink[i]<<=bT; sink[i]|=hPos;   sink[i]<<=bF; sink[i]|=hP1Form;  sink[i]<<=bF; sink[i]|=dForm; ++i; ++t;
//			sink[i]=t; sink[i]<<=bT; sink[i]|=hPos;   sink[i]<<=bF; sink[i]|=hM1Form;  sink[i]<<=bF; sink[i]|=dForm; ++i; ++t;
//			sink[i]=t; sink[i]<<=bT; sink[i]|=hPos;   sink[i]<<=bT; sink[i]|=hP1Pos;   sink[i]<<=bF; sink[i]|=dForm; ++i; ++t;
//			sink[i]=t; sink[i]<<=bT; sink[i]|=hPos;   sink[i]<<=bT; sink[i]|=hM1Pos;   sink[i]<<=bF; sink[i]|=dForm; ++i; ++t;
//
//			sink[i]=t; sink[i]<<=bT; sink[i]|=dPos;   sink[i]<<=bF; sink[i]|=dP1Form;  sink[i]<<=bF; sink[i]|=hForm; ++i; ++t;
//			sink[i]=t; sink[i]<<=bT; sink[i]|=dPos;   sink[i]<<=bF; sink[i]|=dM1Form;  sink[i]<<=bF; sink[i]|=hForm; ++i; ++t;
//			sink[i]=t; sink[i]<<=bT; sink[i]|=dPos;   sink[i]<<=bT; sink[i]|=dP1Pos;   sink[i]<<=bF; sink[i]|=hForm; ++i; ++t;
//			sink[i]=t; sink[i]<<=bT; sink[i]|=dPos;   sink[i]<<=bT; sink[i]|=dM1Pos;   sink[i]<<=bF; sink[i]|=hForm; ++i; ++t;
//		}
//
//		if(Options.x5){
//			sink[i]=t; sink[i]<<=bF; sink[i]|=hForm;   sink[i]<<=bF; sink[i]|=hP1Form;  sink[i]<<=bT; sink[i]|=hM1Pos; sink[i]<<=bT; sink[i]|=dPos; ++i; ++t;
//			sink[i]=t; sink[i]<<=bF; sink[i]|=hForm;   sink[i]<<=bF; sink[i]|=hM1Form;  sink[i]<<=bT; sink[i]|=hP1Pos; sink[i]<<=bT; sink[i]|=dPos; ++i; ++t;
//			sink[i]=t; sink[i]<<=bF; sink[i]|=hForm;   sink[i]<<=bT; sink[i]|=hP1Pos;   sink[i]<<=bT; sink[i]|=hM1Pos; sink[i]<<=bT; sink[i]|=dPos; ++i; ++t;
//
//			sink[i]=t; sink[i]<<=bF; sink[i]|=hForm;   sink[i]<<=bF; sink[i]|=hP1Form;  sink[i]<<=bT; sink[i]|=hM1Pos; sink[i]<<=bT; sink[i]|=dPos; ++i; ++t;
//			sink[i]=t; sink[i]<<=bF; sink[i]|=hForm;   sink[i]<<=bF; sink[i]|=hM1Form;  sink[i]<<=bT; sink[i]|=hP1Pos; sink[i]<<=bT; sink[i]|=dPos; ++i; ++t;
//			sink[i]=t; sink[i]<<=bF; sink[i]|=hForm;   sink[i]<<=bT; sink[i]|=hP1Pos;   sink[i]<<=bT; sink[i]|=hM1Pos; sink[i]<<=bT; sink[i]|=dPos; ++i; ++t;
//
//			sink[i]=t; sink[i]<<=bF; sink[i]|=dForm;   sink[i]<<=bF; sink[i]|=dP1Form;  sink[i]<<=bT; sink[i]|=dM1Pos; sink[i]<<=bT; sink[i]|=hPos; ++i; ++t;
//			sink[i]=t; sink[i]<<=bF; sink[i]|=dForm;   sink[i]<<=bF; sink[i]|=dM1Form;  sink[i]<<=bT; sink[i]|=dP1Pos; sink[i]<<=bT; sink[i]|=hPos; ++i; ++t;
//			sink[i]=t; sink[i]<<=bF; sink[i]|=dForm;   sink[i]<<=bT; sink[i]|=dP1Pos;   sink[i]<<=bT; sink[i]|=dM1Pos; sink[i]<<=bT; sink[i]|=hPos; ++i; ++t;
//
//			sink[i]=t; sink[i]<<=bF; sink[i]|=dForm;   sink[i]<<=bF; sink[i]|=dP1Form;  sink[i]<<=bT; sink[i]|=dM1Pos; sink[i]<<=bT; sink[i]|=hPos; ++i; ++t;
//			sink[i]=t; sink[i]<<=bF; sink[i]|=dForm;   sink[i]<<=bF; sink[i]|=dM1Form;  sink[i]<<=bT; sink[i]|=dP1Pos; sink[i]<<=bT; sink[i]|=hPos; ++i; ++t;
//			sink[i]=t; sink[i]<<=bF; sink[i]|=dForm;   sink[i]<<=bT; sink[i]|=dP1Pos;   sink[i]<<=bT; sink[i]|=dM1Pos; sink[i]<<=bT; sink[i]|=hPos; ++i; ++t;			
//		}
//
//		if(Options.x6){ //Bad for korean (bF+bF+bT+bF needs too many bits)
//			sink[i]=t; sink[i]<<=bF; sink[i]|=hForm;   sink[i]<<=bF; sink[i]|=hP1Form;  sink[i]<<=bT; sink[i]|=hM1Pos; sink[i]<<=bF; sink[i]|=dForm; ++i; ++t;
//			sink[i]=t; sink[i]<<=bF; sink[i]|=hForm;   sink[i]<<=bF; sink[i]|=hM1Form;  sink[i]<<=bT; sink[i]|=hP1Pos; sink[i]<<=bF; sink[i]|=dForm; ++i; ++t;
//			sink[i]=t; sink[i]<<=bF; sink[i]|=hForm;   sink[i]<<=bT; sink[i]|=hP1Pos;   sink[i]<<=bT; sink[i]|=hM1Pos; sink[i]<<=bF; sink[i]|=dForm; ++i; ++t;
//
//			sink[i]=t; sink[i]<<=bF; sink[i]|=hForm;   sink[i]<<=bF; sink[i]|=hP1Form;  sink[i]<<=bT; sink[i]|=hM1Pos; sink[i]<<=bF; sink[i]|=dForm; ++i; ++t;
//			sink[i]=t; sink[i]<<=bF; sink[i]|=hForm;   sink[i]<<=bF; sink[i]|=hM1Form;  sink[i]<<=bT; sink[i]|=hP1Pos; sink[i]<<=bF; sink[i]|=dForm; ++i; ++t;
//			sink[i]=t; sink[i]<<=bF; sink[i]|=hForm;   sink[i]<<=bT; sink[i]|=hP1Pos;   sink[i]<<=bT; sink[i]|=hM1Pos; sink[i]<<=bF; sink[i]|=dForm; ++i; ++t;
//
//			sink[i]=t; sink[i]<<=bF; sink[i]|=dForm;   sink[i]<<=bF; sink[i]|=dP1Form;  sink[i]<<=bT; sink[i]|=dM1Pos; sink[i]<<=bF; sink[i]|=hForm; ++i; ++t;
//			sink[i]=t; sink[i]<<=bF; sink[i]|=dForm;   sink[i]<<=bF; sink[i]|=dM1Form;  sink[i]<<=bT; sink[i]|=dP1Pos; sink[i]<<=bF; sink[i]|=hForm; ++i; ++t;
//			sink[i]=t; sink[i]<<=bF; sink[i]|=dForm;   sink[i]<<=bT; sink[i]|=dP1Pos;   sink[i]<<=bT; sink[i]|=dM1Pos; sink[i]<<=bF; sink[i]|=hForm; ++i; ++t;
//
//			sink[i]=t; sink[i]<<=bF; sink[i]|=dForm;   sink[i]<<=bF; sink[i]|=dP1Form;  sink[i]<<=bT; sink[i]|=dM1Pos; sink[i]<<=bF; sink[i]|=hForm; ++i; ++t;
//			sink[i]=t; sink[i]<<=bF; sink[i]|=dForm;   sink[i]<<=bF; sink[i]|=dM1Form;  sink[i]<<=bT; sink[i]|=dP1Pos; sink[i]<<=bF; sink[i]|=hForm; ++i; ++t;
//			sink[i]=t; sink[i]<<=bF; sink[i]|=dForm;   sink[i]<<=bT; sink[i]|=dP1Pos;   sink[i]<<=bT; sink[i]|=dM1Pos; sink[i]<<=bF; sink[i]|=hForm; ++i; ++t;
//		}
//
//		if(Options.x7){
//			sink[i]=t; sink[i]<<=bT; sink[i]|=hPos;   sink[i]<<=bF; sink[i]|=hP1Form;  sink[i]<<=bT; sink[i]|=hM1Pos; sink[i]<<=bT; sink[i]|=dPos; ++i; ++t;
//			sink[i]=t; sink[i]<<=bT; sink[i]|=hPos;   sink[i]<<=bF; sink[i]|=hM1Form;  sink[i]<<=bT; sink[i]|=hP1Pos; sink[i]<<=bT; sink[i]|=dPos; ++i; ++t;
//			sink[i]=t; sink[i]<<=bT; sink[i]|=hPos;   sink[i]<<=bT; sink[i]|=hP1Pos;   sink[i]<<=bT; sink[i]|=hM1Pos; sink[i]<<=bT; sink[i]|=dPos; ++i; ++t;
//
//			sink[i]=t; sink[i]<<=bT; sink[i]|=hPos;   sink[i]<<=bF; sink[i]|=hP1Form;  sink[i]<<=bT; sink[i]|=hM1Pos; sink[i]<<=bT; sink[i]|=dPos; ++i; ++t;
//			sink[i]=t; sink[i]<<=bT; sink[i]|=hPos;   sink[i]<<=bF; sink[i]|=hM1Form;  sink[i]<<=bT; sink[i]|=hP1Pos; sink[i]<<=bT; sink[i]|=dPos; ++i; ++t;
//			sink[i]=t; sink[i]<<=bT; sink[i]|=hPos;   sink[i]<<=bT; sink[i]|=hP1Pos;   sink[i]<<=bT; sink[i]|=hM1Pos; sink[i]<<=bT; sink[i]|=dPos; ++i; ++t;
//
//			sink[i]=t; sink[i]<<=bT; sink[i]|=dPos;   sink[i]<<=bF; sink[i]|=dP1Form;  sink[i]<<=bT; sink[i]|=dM1Pos; sink[i]<<=bT; sink[i]|=hPos; ++i; ++t;
//			sink[i]=t; sink[i]<<=bT; sink[i]|=dPos;   sink[i]<<=bF; sink[i]|=dM1Form;  sink[i]<<=bT; sink[i]|=dP1Pos; sink[i]<<=bT; sink[i]|=hPos; ++i; ++t;
//			sink[i]=t; sink[i]<<=bT; sink[i]|=dPos;   sink[i]<<=bT; sink[i]|=dP1Pos;   sink[i]<<=bT; sink[i]|=dM1Pos; sink[i]<<=bT; sink[i]|=hPos; ++i; ++t;
//
//			sink[i]=t; sink[i]<<=bT; sink[i]|=dPos;   sink[i]<<=bF; sink[i]|=dP1Form;  sink[i]<<=bT; sink[i]|=dM1Pos; sink[i]<<=bT; sink[i]|=hPos; ++i; ++t;
//			sink[i]=t; sink[i]<<=bT; sink[i]|=dPos;   sink[i]<<=bF; sink[i]|=dM1Form;  sink[i]<<=bT; sink[i]|=dP1Pos; sink[i]<<=bT; sink[i]|=hPos; ++i; ++t;
//			sink[i]=t; sink[i]<<=bT; sink[i]|=dPos;   sink[i]<<=bT; sink[i]|=dP1Pos;   sink[i]<<=bT; sink[i]|=dM1Pos; sink[i]<<=bT; sink[i]|=hPos; ++i; ++t;
//
//		}
//
//
//
//		if(Options.x8){
//
//			sink[i]=t; sink[i]<<=bT; sink[i]|=hPos;   sink[i]<<=bF; sink[i]|=hP1Form;  sink[i]<<=bT; sink[i]|=hM1Pos; sink[i]<<=bF; sink[i]|=dForm; ++i; ++t;
//			sink[i]=t; sink[i]<<=bT; sink[i]|=hPos;   sink[i]<<=bF; sink[i]|=hM1Form;  sink[i]<<=bT; sink[i]|=hP1Pos; sink[i]<<=bF; sink[i]|=dForm; ++i; ++t;
//			sink[i]=t; sink[i]<<=bT; sink[i]|=hPos;   sink[i]<<=bT; sink[i]|=hP1Pos;   sink[i]<<=bT; sink[i]|=hM1Pos; sink[i]<<=bF; sink[i]|=dForm; ++i; ++t;
//
//			sink[i]=t; sink[i]<<=bT; sink[i]|=hPos;   sink[i]<<=bF; sink[i]|=hP1Form;  sink[i]<<=bT; sink[i]|=hM1Pos; sink[i]<<=bF; sink[i]|=dForm; ++i; ++t;
//			sink[i]=t; sink[i]<<=bT; sink[i]|=hPos;   sink[i]<<=bF; sink[i]|=hM1Form;  sink[i]<<=bT; sink[i]|=hP1Pos; sink[i]<<=bF; sink[i]|=dForm; ++i; ++t;
//			sink[i]=t; sink[i]<<=bT; sink[i]|=hPos;   sink[i]<<=bT; sink[i]|=hP1Pos;   sink[i]<<=bT; sink[i]|=hM1Pos; sink[i]<<=bF; sink[i]|=dForm; ++i; ++t;
//
//
//
//			sink[i]=t; sink[i]<<=bT; sink[i]|=dPos;   sink[i]<<=bF; sink[i]|=dP1Form;  sink[i]<<=bT; sink[i]|=dM1Pos; sink[i]<<=bF; sink[i]|=hForm; ++i; ++t;
//			sink[i]=t; sink[i]<<=bT; sink[i]|=dPos;   sink[i]<<=bF; sink[i]|=dM1Form;  sink[i]<<=bT; sink[i]|=dP1Pos; sink[i]<<=bF; sink[i]|=hForm; ++i; ++t;
//			sink[i]=t; sink[i]<<=bT; sink[i]|=dPos;   sink[i]<<=bT; sink[i]|=dP1Pos;   sink[i]<<=bT; sink[i]|=dM1Pos; sink[i]<<=bF; sink[i]|=hForm; ++i; ++t;
//
//			sink[i]=t; sink[i]<<=bT; sink[i]|=dPos;   sink[i]<<=bF; sink[i]|=dP1Form;  sink[i]<<=bT; sink[i]|=dM1Pos; sink[i]<<=bF; sink[i]|=hForm; ++i; ++t;
//			sink[i]=t; sink[i]<<=bT; sink[i]|=dPos;   sink[i]<<=bF; sink[i]|=dM1Form;  sink[i]<<=bT; sink[i]|=dP1Pos; sink[i]<<=bF; sink[i]|=hForm; ++i; ++t;
//			sink[i]=t; sink[i]<<=bT; sink[i]|=dPos;   sink[i]<<=bT; sink[i]|=dP1Pos;   sink[i]<<=bT; sink[i]|=dM1Pos; sink[i]<<=bF; sink[i]|=hForm; ++i; ++t;
//		}
		return i;
	}

}
