package ims.io;

import ims.data.CoNLL06Sentence;
import ims.data.CoNLL09Sentence;
import ims.data.ISentence;
import ims.data.NBestCoNLL09Sentence;
import ims.data.StagStackSentence;
import ims.util.AbstractSingleItemBufferedIterator;
import ims.util.IOUtils;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

public class TabbedReader<T extends ISentence> extends AbstractReader<T> {

	private final File input;
	private final Class<T> sentenceClass;
	
	public TabbedReader(File input,Class<T> sentenceClass){
		this.input=input;
		this.sentenceClass=sentenceClass;
	}
	
	@Override
	public Iterator<T> iterator() {
		BufferedReader br=IOUtils.getBufferedReader(input);
		try {
			return new TabbedSentenceIterator<T>(br);
		} catch (IOException e) {
			e.printStackTrace();
		} catch (Exception e){
			e.printStackTrace();
			System.err.println();
			System.err.println();
			System.err.println();
			System.err.println("Caught exception when initializing sentence iterator -- ?wrong input format?");
		}
		return null;
	}
	
	class TabbedSentenceIterator<S> extends AbstractSingleItemBufferedIterator<T> {

		private final BufferedReader br;
		int sId=0;

		TabbedSentenceIterator(BufferedReader br) throws IOException{
			this.br=br;
			super.initSetNext(init());
		}
		
		@Override
		protected T readNext() throws IOException {
			List<String> lines=new ArrayList<String>();
			String line;
			while((line=br.readLine())!=null){
				if(line.isEmpty() || line.matches(EMPTY_LINE_REGEXP))
					break;
				lines.add(line);
			}
			if(lines.isEmpty()){
				br.close();
				return null;
			}
			
			return lines2Sentence(lines,sId++,sentenceClass);
		}

		@Override
		protected T init() throws IOException {
			return readNext();
		}
		
	}

	
	//XXX better get things right here
	@SuppressWarnings("unchecked")
	public static <T extends ISentence> T lines2Sentence(List<String> lines,int sId,Class<T> sentenceClass) throws Error {
		ISentence s;
		if(sentenceClass==CoNLL06Sentence.class){
			s=CoNLL06Sentence.lines2Sentence(lines, sId);
		} else if(sentenceClass==CoNLL09Sentence.class){
			s=CoNLL09Sentence.lines2Sentence(lines, sId);
		} else if(sentenceClass==StagStackSentence.class){
			s=StagStackSentence.lines2Sentence(lines, sId);
		} else if(sentenceClass==NBestCoNLL09Sentence.class){
			s=NBestCoNLL09Sentence.lines2Sentence(lines, sId);
		} else {
			throw new Error("unknown format");
		}
		return (T) s;
	}

}
