package ims.oracle;

import ims.oracle.Configuration.TR;
import ims.trans.GoldDynParser;
import ims.trans.Model;
import ims.trans.Options;
import ims.trans.Options.Oracle;
import ims.trans.Options.TransitionSystem;
import ims.util.ZipUtil;

public class OracleFactory {
	
	public static IOracle getOracle(Options options){
		Oracle o=options.oracle;
		final IOracle oracle;
		if(o==Oracle.File){
			oracle=DeterministicFileOracle.create(options);
		} else if(o==Oracle.Model){
			Model m=null;
			try {
				m = ZipUtil.loadObjectFromEntry(Model.ZIP_FILE_ENTRY, Options.goldModel, Model.class);
			} catch (Exception e) {
				e.printStackTrace();
				System.exit(1);
			}
			GoldDynParser gdp=new GoldDynParser(m,Options.goldDecodePolicy);
			oracle=new DeterministicModelOracle(gdp,Options.goldBeam);
		} else if(o==Oracle.Static || o==Oracle.Dynamic){
			return getVanillaOracle(o,options.transitionSystem);
		} else {
			throw new Error("not implemented: "+o+" "+options.transitionSystem);
		}
		return oracle;
	}
	
	public static IOracle getVanillaOracle(Oracle o,TransitionSystem ts){
		if(o==Oracle.Dynamic)
			return getDynamicOracle(ts);
		else if(o==Oracle.Static)
			return getStaticOracle(ts);
		else
			throw new Error("you are wrong here");
	}

	public static AbstractStaticOracle getStaticOracle(TransitionSystem ts) {
		switch(ts){
		case ArcS:            return new DeterministicArcStandardOracle(ts.trs);
		case ArcSLazy:        return new DeterministicArcStandardLazyOracle(ts.trs);
		case ArcSOldSwap:     return new DeterministicArcStandardSwapOldOracle(ts.trs);
		case ArcSNewSwap:     return new DeterministicArcStandardSwapNewOracle(ts.trs);
		case ArcSOptimalSwap: return new DeterministicArcStandardOptimalSwapOracle(ts.trs);
		default:              throw new Error("unknown static oracle: "+ts);
		}
	}
	
	public static AbstractDynamicOracle getDynamicOracle(TransitionSystem ts){
		switch(ts){
		case ArcS:            return new DynamicArcStandardOracle(ts.trs);
		case ArcSDynSwap:     return new DynamicArcStandardSwapOracle(ts.trs);
		case ArcSFullDynSwap: return new DynamicArcStandardSwapFullOracle(ts.trs);
		case ArcSLASWDynSwap: return new DynamicArcStandardSwapLASWDynOracle(ts.trs);
		case ArcSConLASWDynSwap: return new DynamicArcStandardSwapConLASWDynOracle(ts.trs);
		case ArcSSwapDynDelay:   return new DelayDynamicOracle(ts.trs);
		case ArcSDynDelay:       return new DelayDynamicOracle(ts.trs);
		default:              throw new Error("unknown dynamic oracle: "+ts);
		}
	}

	public static IOracle getDynSentOracle() {
		return new DynamicSentenceSplittingOracle(new TR[]{TR.LA,TR.RAS,TR.SW,TR.SH,TR.SB});
	}
	
}
