package ims.tag;

import ims.hotcoref.symbols.TagSymbolTable;
import ims.perceptron.FV3;
import ims.perceptron.IFV;
import ims.tag.features.AbstractFeatureSet;
import ims.tag.features.FeatureScorer;
import ims.tag.features.Potentials;

public class GreedyLocalDecoder implements IDecoder {
	private static final long serialVersionUID = 1664072132238618192L;
	
	private final FeatureScorer featureScorer;
	
	public GreedyLocalDecoder(FeatureScorer featureScorer){
		this.featureScorer=featureScorer;
	}
	
	@Override
	public int[] decodeInstance(TagInstance instance, TagSymbolTable symTab) {
		Potentials potentials=featureScorer.getPotentials(instance);
		int[] r = greedyFromPotentials(instance, potentials);
		return r;
	}
	@Override
	public double recoverFeaturesAndFillFV(int[] tags, TagInstance inst, AbstractFeatureSet fs, IFV<FV3> fv, FeatureScorer featureScorer) {
		TagContext tagContext=null;
		fv.clear();
		double score=0.d;
		for(int i=1;i<tags.length;++i){
			score+=fs.recoverFeaturesAndFillFV(inst, tagContext, i, tags[i], featureScorer, fv);
		}
		return score;
	}

	public static int[] greedyFromPotentials(TagInstance instance, Potentials potentials) {
		int[] r=new int[instance.forms.length];
		for(int idx=1;idx<r.length;++idx){
			r[idx]=potentials.bestPotential(idx);
		}
		return r;
	}

}
