package ims.tag.features;

import java.util.Arrays;

import ims.hotcoref.symbols.TagSymbolTable;
import ims.perceptron.Long2IntInterface;
import ims.tag.ITagContext;
import ims.tag.TagContext;
import ims.tag.TagInstance;
import ims.util.ArrayFunctions;

public class FeatureScorer {

	final AbstractFeatureSet fs;
	final Long2IntInterface l2i;
	final float[] parameters;
	final int bT;
	final int[] tags;
	final int tagLen;
	
	public FeatureScorer(AbstractFeatureSet fs,Long2IntInterface l2i,float[] params,TagSymbolTable symTab){
		this.fs=fs;
		this.l2i=l2i;
		this.parameters=params;
		this.bT=symTab.tags.getBits();
		this.tagLen=symTab.tags.getItems();
		this.tags=Arrays.copyOf(symTab.tagIdx, symTab.tagIdx.length);
	}
	
	public Potentials getPotentials(TagInstance inst){
		double[][] potentials=new double[inst.forms.length][tagLen];
		for(int i=1;i<inst.forms.length;++i){
			int m=fs.extractLocalFeatures(i, inst);
			ArrayFunctions.shiftAllLeftByBits(fs.sink, bT, m);
			for(int t:tags){
				double score=0.d;
				for(int q=0;q<m;++q){
					long l=fs.sink[q];
					l|=t;
					int y=l2i.l2i(l);
					score+=parameters[y];
				}
				potentials[i][t]=score;
			}
		}
		return new Potentials(potentials);
	}
	
	public double scoreNonLocal(int idx,TagInstance inst,ITagContext context,int tag){
		int m=fs.extractNonLocalFeatures(idx, inst, context);
		ArrayFunctions.shiftAllLeftByBits(fs.sink, bT, m);
		double score=0.d;
		for(int i=0;i<m;++i){
			long l=fs.sink[i];
			l|=tag;
			int y=l2i.l2i(l);
			score+=parameters[y];
		}
		return score;
	}
	
	public double[] scoreNonLocal(int idx,TagInstance inst,TagContext context){
		double[] r=new double[tagLen];
		int m=fs.extractNonLocalFeatures(idx, inst, context);
		ArrayFunctions.shiftAllLeftByBits(fs.sink, bT, m);
		for(int t:tags){
			double score=0.d;
			for(int i=0;i<m;++i){
				long l=fs.sink[i];
				l|=t;
				int y=l2i.l2i(l);
				score+=parameters[y];
			}
			r[t]=score;
		}
		return r;
	}
	
	public double scoreFeatures(long[] sink,int m){
		double s=0.d;
		for(int i=0;i<m;++i){
			int y=l2i.l2i(sink[i]);
			s+=parameters[y];
		}
		return s;
	}
	
}
