package ims.trans;

import java.io.IOException;

import ims.data.StagStackSentence;
import ims.io.ReaderWriterFactory;
import ims.io.SentenceReader;
import ims.io.SentenceWriter;
import ims.trans.util.Eval;
import ims.util.Accuracy;
import ims.util.DBO;
import ims.util.Pair;
import ims.util.ZipUtil;

public class AutoStackTest {

	public static void main(String[] args) throws ClassNotFoundException, IOException{
		Options options=new Options(args);
		DBO.println("Loading model1");
		Model m1=ZipUtil.loadObjectFromEntry(Model.ZIP_FILE_ENTRY, options.model, Model.class);
		DBO.println("Done.");
		DBO.println("Loading model2");
		Model m2=ZipUtil.loadObjectFromEntry(Model.ZIP_FILE_ENTRY, options.model2, Model.class);
		DBO.println("Done.");

		AbstractDecoder d1=new Decoder(m1);
		AbstractDecoder d2=new Decoder(m2);

		//We pass false for both reversing, and then we do the reversing manually in the parse loop
		SentenceReader<StagStackSentence> reader=ReaderWriterFactory.getReader(options.test, StagStackSentence.class);
		SentenceWriter writer=ReaderWriterFactory.getWriter(options.out,false);
		DBO.println("Parsing");
		DBO.printNoPrefix("Sentence: ");
		int erase=DBO.eraseAndPrint(0, 0);
		int sentCount=0;
		long start=System.currentTimeMillis();
		for(StagStackSentence s:reader){
			{ //L1
				//Reverse if necessary 
				if(m1.reversed)
					s.reverseSentence();
				//Parse L1
				State state=d1.parse(s);
				Pair<int[],String[]> p=d1.state2tree(state);
				int[] heads=p.getLeft();
				String[] lbl=p.getRight();

				//Set stack information
				s.setStackHead(heads);
				s.setStackLbl(lbl);
				//Reverse back if applicable
				if(m1.reversed)
					s.reverseSentence();
			}
			//
			// Done with L1
			//
			{ //L2
				//Reverse if necessary
				if(m2.reversed)
					s.reverseSentence();
				//Parse L2
				State state=d2.parse(s);
				Pair<int[],String[]> p=d2.state2tree(state);
				int[] heads=p.getLeft();
				String[] lbl=p.getRight();
				s.setPredHeads(heads);
				s.setPredLabels(lbl);
				
				//Reverse back if applicable
				if(m2.reversed)
					s.reverseSentence();				
			}
			writer.write(s);
			++sentCount;
			if(sentCount%100==0){
				erase=DBO.eraseAndPrint(erase, String.format("%d (%dms/sentence)", sentCount,(System.currentTimeMillis()-start)/sentCount));
			}
		}
		DBO.println();
		DBO.println();
		writer.close();
		if(options.gold!=null){
			Pair<Accuracy,Accuracy> pair=Eval.getUASandLAS(options.out, options.gold, options.format);
			System.out.println("UAS: "+pair.getLeft().toString() + "   LAS: "+pair.getRight().toString());
		}
		options.done();
	}
	
}
