package ims.trans;

import ims.data.ISentence;
import ims.io.ReaderWriterFactory;
import ims.io.SentenceReader;
import ims.oracle.IOracle;
import ims.oracle.OracleFactory;
import ims.trans.util.FileTRSeq;
import ims.util.DBO;
import ims.util.Util;
import ims.util.ZipUtil;

import java.io.IOException;
import java.io.Serializable;
import java.util.List;

public class CVPrintTRSeq {

	public static void main(String[] args) throws ClassNotFoundException, IOException{
		Options options=new Options(args);
		DBO.setOut(System.err);
		DBO.println("Loading models");
		CVStruct cvModel=ZipUtil.loadObjectFromEntry(CVStruct.ZIP_FILE_ENTRY,options.model,CVStruct.class);
		DBO.println("Done.");
		
		SentenceReader<?> reader=ReaderWriterFactory.getReader(options.train, options.format.sentenceClass);

		IOracle oracle=OracleFactory.getVanillaOracle(options.oracle, options.transitionSystem);
		
		int f=0;
		Decoder d=getDecoder(f,cvModel);
		
		int i=0;
		for(ISentence s:reader){
			if(s.getLength()!=cvModel.senLengths[i]){
				System.err.printf("Length mismatch on sentence %d. Model: %d, File: %d\n",i,cvModel.senLengths[i],s.getLength());
				System.err.println("Exiting!");
				System.exit(1);
			}
			
			State terminalState=d.decodeGold(s, options.beam, oracle);
			
			List<String> trSeq=FileTRSeq.extractSeqStringList(terminalState, d.symTab);
			
			System.out.println(Util.joinStrings(trSeq, ","));
			
			
			++i;
			if(i==cvModel.senLengths[f] && f+1<cvModel.folds){
				++f;
				d=getDecoder(f,cvModel);
			}
		}
		
		options.done();
	}
	
	
	
	
	
	private static Decoder getDecoder(int f, CVStruct cvModel) {
		return new Decoder(cvModel.models.get(f));
	}





	public static class CVStruct implements Serializable {
		private static final long serialVersionUID = -6671801567929663399L;
		public static final String ZIP_FILE_ENTRY="cvShareModel";
		public final int folds;
		public final int[] senLengths;
		public final int[] sOffsets;
		public final List<Model> models;
		public CVStruct(int folds,int[] senLengths,int[] sOffsets,List<Model> models) {
			this.folds = folds;
			this.senLengths = senLengths;
			this.sOffsets = sOffsets;
			this.models = models;
		}
	}
}
