package ims.trans.transitions;

import ims.data.Instance;
import ims.hotcoref.symbols.SymbolTable;
import ims.oracle.Configuration.TR;
import ims.trans.State;

public class DefaultDocBlockingSBSwapStandardFilter extends DefaultArcStandardSwapTransitionFilter {
	private static final long serialVersionUID = 3692965612695507051L;

	

	final int[] SH_LA_RA_SW_BSB;
	final int[] SH_LA_RA_BSB;
	final int[] LA_RA_SW_BSB;
	final int[] LA_RA_BSB;
	final int[] SH_RA_BSB;
	final int[] SH_BSB;
	final int[] RA_BSB;
	
	public DefaultDocBlockingSBSwapStandardFilter(SymbolTable symTab) {
		super(symTab);
		int leftArcs=0;
		int rightArcs=0;
		
		for(int i=2;i<symTab.transition2label.length;++i){
			ITransition tr=symTab.transition2label[i];
			switch(tr.getTR()){
			case RAS: ++rightArcs; break;
			case LA:  ++leftArcs;  break;
			default:  //do nothing
			}
		}
		
		SH_LA_RA_SW_BSB=new int[3+leftArcs+rightArcs];         int a=0;
		SH_LA_RA_BSB=new int[2+leftArcs+rightArcs];            int b=0;
		LA_RA_SW_BSB=new int[2+leftArcs+rightArcs];            int c=0;
		LA_RA_BSB=new int[1+leftArcs+rightArcs];               int d=0;
		SH_RA_BSB=new int[2+rightArcs];                        int e=0;
		SH_BSB=new int[2];                                     int f=0;
		RA_BSB=new int[1+rightArcs];                           int g=0;
		
		for(int i=2;i<symTab.transition2label.length;++i){
			ITransition tr=symTab.transition2label[i];
			TR t=tr.getTR();
			int id=tr.getId();
			SH_LA_RA_SW_BSB[a++]=id; //This one has them all
			boolean inB=false,inC=false,inD=false,inE=false,inF=false,inG=false;
			switch(t){
			case SB: inB=true; inC=true; inD=true; inE=true; inF=true; inG=true; break;
			case LA:  inB=true; inC=true; inD=true;                               break;
			case RAS: inB=true; inC=true; inD=true; inE=true; inG=true;           break;
			case SH:  inB=true; inE=true; inF=true;                               break;
			case SW:  inC=true;                                                   break;
			default:  throw new Error("this shouldn't happen");
			}
			if(inB)
				SH_LA_RA_BSB[b++]=id;
			if(inC)
				LA_RA_SW_BSB[c++]=id;
			if(inD)
				LA_RA_BSB[d++]=id;
			if(inE)
				SH_RA_BSB[e++]=id;
			if(inF)
				SH_BSB[f++]=id;
			if(inG)
				RA_BSB[g++]=id;
		}
		if(a!=SH_LA_RA_SW_BSB.length ||
				b!=SH_LA_RA_BSB.length ||
				c!=LA_RA_SW_BSB.length ||
				d!=LA_RA_BSB.length ||
				e!=SH_RA_BSB.length ||
				f!=SH_BSB.length ||
				g!=RA_BSB.length)
			throw new Error("length mismatch");
	}
	
	@Override
	public int[] getAllowedTransitions(State s, Instance inst) {
		boolean twoItemsOnStack=s.tail!=null;
		boolean shiftBlocked=s.isBlockedForShift();
		
		if(shiftBlocked && !twoItemsOnStack)
			throw new Error("Shift is blocked but there is only one item on stack. Doesn't make sense.");
		
		boolean canBSB=s.sbPermissible();
		
		if(!twoItemsOnStack){
			if(canBSB)
				return SH_BSB;
			else
				return SH;
		}
		
		//Can always rightarc
		boolean canLeftArc=s.tail.s0!=0; //can leftarc if s1 is not root
		boolean canSwap=s.tail.s0 < s.s0 && canLeftArc; //can't swap (two items that we're already swapped), (if s1 is root)
		boolean canShift=s.buf.get(0)>0 && !shiftBlocked;
		
		if(canBSB){
			if(canLeftArc && canSwap && canShift)
				return SH_LA_RA_SW_BSB;
			if(canLeftArc && canSwap)
				return LA_RA_SW_BSB;
			if(canLeftArc && canShift)
				return SH_LA_RA_BSB;
			if(canSwap && canShift)
				throw new Error("this cannot happen SINCE canSwap but not canLeftArc");
			if(canSwap)
				throw new Error("this cannot happen SINCE canSwap but not canLeftArc");
			if(canLeftArc)
				return LA_RA_BSB;
			if(canShift)
				return SH_RA_BSB;
			
			return RA_BSB;
		} else {
			if(canLeftArc && canSwap && canShift)
				return SH_SW_LA_RA;
			if(canLeftArc && canShift)
				return SH_LA_RA;
			if(canSwap && canShift)
				throw new Error("this cannot happen SINCE canSwap but not canLeftArc");
			if(canLeftArc && canSwap)
				return SW_LA_RA;
			if(canSwap)
				throw new Error("this cannot happen SINCE canSwap but not canLeftArc");
			if(canLeftArc)
				return LA_RA;
			if(canShift)
				return SH_RA;
			
			return RA;
		}
	}

	@Override
	public boolean isPOSFiltered() {
		return false;
	}

}
