package ims.trans.util;

import ims.hotcoref.symbols.SymbolTable;
import ims.trans.State;
import ims.util.AbstractSingleItemBufferedIterator;
import ims.util.IOUtils;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

public class FileTRSeq {

	public static List<String> extractSeqStringList(State state,SymbolTable symTab){
		List<String> r=new ArrayList<String>();
		State c=state;
		while(c.lastAction!=null){
			String tr=symTab.transition2label[c.lastAction.getId()].getFullTransitionLabel();
			r.add(tr);
			c=c.previous;
		}
		
		Collections.reverse(r);
		
		return r;
	}
	
	public static int[] parseFromFile(String line,SymbolTable symTab){
		String[] actions=line.split(",");
		int[] r=new int[actions.length];
		for(int i=0;i<actions.length;++i){
			int a=symTab.transitions.lookup(actions[i]);
			r[i]=a;
		}
		return r;
	}
	
	public static List<int[]> parseFromFile(File f,SymbolTable symTab) throws IOException{
		List<int[]> r=new ArrayList<int[]>();
		BufferedReader reader=IOUtils.getBufferedReader(f);
		String line;
		while((line=reader.readLine())!=null){
			r.add(parseFromFile(line,symTab));
		}
		return r;
	}
	
	public static Iterator<List<String>> getSeqIt(File f) throws IOException{
		BufferedReader br=IOUtils.getBufferedReader(f);
		return new SeqIt(br);
	}
	
	public static final class SeqIt extends AbstractSingleItemBufferedIterator<List<String>> {

		final BufferedReader br;
		
		SeqIt(BufferedReader br) throws IOException{
			this.br=br;
			super.initSetNext(init());
		}
		
		@Override
		protected List<String> readNext() throws IOException {
			String l=br.readLine();
			if(l==null){
				br.close();
				return null;
			}
			List<String> r=Arrays.asList(l.split(","));
			return r;
		}

		@Override
		protected List<String> init() throws IOException {
			return readNext();
		}
		
	}
}
