package ims.util;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.RandomAccessFile;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.List;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

public class IOUtils {

	public static final String DEFAULT_ENCODING="UTF-8";
	
	public static BufferedReader getSTDINBufferedReader() throws UnsupportedEncodingException{
		return getSTDINBufferedReader(DEFAULT_ENCODING);
	}
	
	public static BufferedReader getSTDINBufferedReader(String encoding) throws UnsupportedEncodingException{
		BufferedReader br=new BufferedReader(new InputStreamReader(System.in,encoding));
		return br;
	}
	
	
	public static BufferedReader getBufferedReader(File f){
		return getBufferedReader(f,DEFAULT_ENCODING,isGZipped(f));
	}
	public static BufferedReader getBufferedReader(File f,String enc,boolean gzipped){
		try {
			InputStream is=new FileInputStream(f.toString());
			if(gzipped)
				is=new GZIPInputStream(is);
			BufferedReader br=new BufferedReader(new InputStreamReader(is,enc));
			return br;
		} catch(IOException e){
			throw new RuntimeException(e);
		}
	}
	
	public static BufferedWriter getBufferedWriter(File f){
		return getBufferedWriter(f,DEFAULT_ENCODING,f.toString().endsWith(".gz"));
	}
	
	public static BufferedWriter getBufferedWriter(File f,String enc,boolean gzipped){
		try {
			OutputStream os=new FileOutputStream(f);
			if(gzipped)
				os=new GZIPOutputStream(os);
			BufferedWriter bw=new BufferedWriter(new OutputStreamWriter(os,enc));
			return bw;
		} catch(IOException e){
			throw new RuntimeException(e);
		}
	}
	
	public static boolean isGZipped(File f) {
		int magic = 0;
		try {
			RandomAccessFile raf = new RandomAccessFile(f, "r");
			magic = raf.read() & 0xff | ((raf.read() << 8) & 0xff00);
			raf.close();
		} catch (Throwable e) {
			e.printStackTrace(System.err);
		}
		return magic == GZIPInputStream.GZIP_MAGIC;
	}
	
	public static boolean createParentFolderIfNecessary(String path){
		File parent=new File(path).getParentFile().getAbsoluteFile();
		if(!parent.exists()){
			return parent.mkdirs();
		}
		if(parent.isDirectory())
			return true;
		else
			return false;
	}
	
	public static List<String> slurpLines(BufferedReader reader) throws IOException{
		List<String> r=new ArrayList<String>();
		String l;
		while((l=reader.readLine())!=null){
			r.add(l);
		}
		reader.close();
		return r;
	}
}
