package ims.util;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Random;

public class JackKnifeSplit {
	
	
	private static final long RAND_SEED=23562346231l;
	private static int SPLITS=10;
	private static boolean SHUFFLE=false;
	
	public static void main(String[] args) throws Exception{
		String input=args[0];
		String outputPrefix=args[1];
		//Parse additional cmd line arguments. Either -splits <i> or -shuffle
		for(int ai=2;ai<args.length;++ai){
			if(args[ai].equalsIgnoreCase("-shuffle")){
				SHUFFLE=true;
			} else if(args[ai].equals("-splits")){
				SPLITS=Integer.parseInt(args[++ai]);
			} else {
				System.err.println("Unknown option: "+args[ai]);
				System.exit(1);
			}
		}
		//Count sentences
		BufferedReader r=new BufferedReader(new InputStreamReader(new FileInputStream(input),"utf8"));
		int sentCount=0;
		List<Integer> senIDs=new ArrayList<Integer>();
		String line;
		while((line=r.readLine())!=null){
			if(line.matches("^\\s*$")) //Is new line
				senIDs.add(sentCount++);
		}
		r.close();
//		System.out.println("Total sentences: "+sentCount);
		//Shuffle?
		if(SHUFFLE){
			Random rand=new Random(RAND_SEED);
			Collections.shuffle(senIDs, rand);
		}
		//Now decide which sentence goes into which chunk
		int[] toSplit=new int[sentCount];
		int n=0;
		for(int m=sentCount/SPLITS,s=0;n<sentCount && s<SPLITS;n=m,m+=sentCount/SPLITS,++s){
//			System.out.printf("n: %d, m: %d, s: %d\n",n,m,s);
			for(int i=n;i<m;++i){
				toSplit[senIDs.get(i)]=s;
			}
		}
		//Handle the last couple of sentences when it can't be evenly divided by the number of splits
		for(;n<sentCount;++n){
			toSplit[senIDs.get(n)]=SPLITS-1;
		}
//		System.exit(0);
		//Dump the split
//		for(int i=0;i<toSplit.length;++i){
//			System.out.printf("%5d   %d\n",i,toSplit[i]);
//		}
		//Setup output writers
		BufferedWriter[] trainWriters=new BufferedWriter[SPLITS];
		BufferedWriter[] testWriters=new BufferedWriter[SPLITS];
		BufferedWriter allWriter=new BufferedWriter(new OutputStreamWriter(new FileOutputStream(outputPrefix+"-train-all"),"utf8"));
		for(int i=0;i<SPLITS;++i){
			trainWriters[i]=new BufferedWriter(new OutputStreamWriter(new FileOutputStream(outputPrefix+"-train-"+i),"utf8"));
			testWriters[i]=new BufferedWriter(new OutputStreamWriter(new FileOutputStream(outputPrefix+"-test-"+i),"utf8"));
		}
		//Write out
		BufferedReader reader=new BufferedReader(new InputStreamReader(new FileInputStream(input),"utf8"));
		List<String> buffer=new ArrayList<String>();
		int t=0;
		while((line=reader.readLine())!=null){
			if(line.matches("^\\s*$")){
				int s=toSplit[t];
				//Print it out
				writeOut(trainWriters, testWriters, buffer, s,allWriter);
				buffer.clear();
				++t;
			} else {
				buffer.add(line);
			}
		}
		//If it doesn't end with new line and we now have stuff left in the buffer, just dump it in the last split
		if(!buffer.isEmpty()) 
			writeOut(trainWriters,testWriters,buffer,SPLITS-1,allWriter);
		reader.close();
		//Close writers
		for(int i=0;i<SPLITS;++i){
			trainWriters[i].close();
			testWriters[i].close();
		}
		allWriter.close();
	}

	private static void writeOut(BufferedWriter[] trainWriters, BufferedWriter[] testWriters, List<String> buffer, int s, BufferedWriter allWriter) throws IOException {
		for(int j=0;j<SPLITS;++j){
			if(j==s){ //Goes in this split
				writeSentence(testWriters[j], buffer);
			} else { //Otherwise in all other training splits
				writeSentence(trainWriters[j], buffer);
			}
		}
		writeSentence(allWriter,buffer); //and goes to all
	}

	private static void writeSentence(BufferedWriter writer,List<String> buffer) throws IOException {
		for(String l:buffer){
			writer.write(l);
			writer.newLine();
		}
		writer.newLine();
	}

}
